/*
 * game server
 */
import java.net.*;
import java.util.*;
import java.io.*;

public class ShootingForDummiesServer implements Runnable {
    // hash of ships on "id" (ie: ip & port).
    static Hashtable clients = new Hashtable();
    static pList plasma = new pList();
    static DatagramSocket sock = null;

    // this is the game logic thread.
    public void run(){
        for(;;){
            StringBuffer sb = new StringBuffer();

            long tim = System.currentTimeMillis();

            // remove "old" plasma (>2 seconds old).
            while(plasma.size() > 0){
                Plasma p = (Plasma)plasma.get(0);
                if(tim - p.tim < 2000)
                    break;
                plasma.shift();     // remove plasma
            }

            Enumeration cliIds = clients.keys();
            // loop for all clients
            while(cliIds.hasMoreElements()){
                Ship s = (Ship)clients.get(cliIds);
                //  update locations
                s.x += Math.cos(s.direction)*s.speed;
                s.y += Math.sin(s.direction)*s.speed;
                // crap around screen clipping
                if(s.x > 1)
                    s.x = -1;
                if(s.x < -1)
                    s.x = 1;
                if(s.y > 1)
                    s.y = -1;
                if(s.y < -1)
                    s.y = 1;
                //  detect plasma collisions
                for(int i=0;i<plasma.size();i++){
                    Plasma p = (Plasma)plasma.get(i);
                    double dist = Math.sqrt(  
                        (s.x - p.x)*(s.x - p.x) +
                        (s.y - p.y)*(s.y - p.y)
                    );
                    if(dist < 1.0/50.0){
                        // find shooting player 
                        Ship shooter = (Ship)clients.get(p.id);
                        // update score.
                        if(shooter != null){
                            shooter.score++;
                        }
                    }
                }
                
                //  shoot plasma (TODO).
                if(s.keys[' ']){
                    if(tim - s.shootTim > 250){
                        s.shootTim = tim;
                        Plasma p = new Plasma();
                        p.id = s.id;
                        p.x = s.x + Math.cos(s.direction);
                        p.y = s.y + Math.sin(s.direction);
                        p.direction = s.direction;
                        p.tim = tim;    // shot time.
                        plasma.push(p);
                    }
                }
                
                // handle user inputs.
                if(s.keys[1006] || s.keys['a']){ // left key
                    s.direction += Math.PI/32;
                }

                if(s.keys[1007] || s.keys['d']){ // right key
                    s.direction -= Math.PI/32;            
                }

                if(s.keys[1004] || s.keys['w']){ // up key
                    s.speed += 0.001;
                    s.speed = Math.min(0.010,s.speed);
                }
                if(s.keys[1005] || s.keys['s']){ // down key
                    s.speed -= 0.002;
                    s.speed = Math.max(-0.005,s.speed);
                }

                if(tim - s.tim > 5*60*1000){
                    clients.remove(cliIds);
                    continue;
                }

                sb.append(""+
                    s.name+","+
                    s.color+","+
                    s.x+","+
                    s.y+","+
                    s.direction+","+
                    s.speed+","+
                    s.score+"\n");
            }
            for(int i=0;i<plasma.size();i++){
                Plasma p = (Plasma)plasma.get(i);
                
                p.x += Math.cos(p.direction)*0.015;
                p.y += Math.sin(p.direction)*0.015;
                // crap around screen clipping
                if(p.x > 1)
                    p.x = -1;
                if(p.x < -1)
                    p.x = 1;
                if(p.y > 1)
                    p.y = -1;
                if(p.y < -1)
                    p.y = 1;

                sb.append(""+
                    "#,"+
                    ","+
                    p.x+","+
                    p.y+","+
                    p.direction+","+","+"\n");
            }

            // notify all clients of new state
            cliIds = clients.keys();
            String msgToAll = sb.toString();
            byte[] ourOutData = msgToAll.getBytes();

            // loop for all clients
            while(cliIds.hasMoreElements()){
                Ship s = (Ship)clients.get(cliIds);
                // create a package.
                DatagramPacket packetOut = new DatagramPacket(ourOutData, ourOutData.length,s.ip,s.port);
                try {
                    sock.send(packetOut);
                } catch (IOException e){
                    // do nothing.
                }
            }
            try{
                Thread.currentThread().sleep(30);
            }catch(Exception e){
                // sleep interrupted (urgh!)
            }
        }
    }

    // main server thread (get msgs from clients).
    public static void main(String[] args) throws Throwable {

        System.out.println("Starting...");

        // loop for each command line argument.
        for(int i=0;i<args.length;i++){
            // split each one on =, ie: name=value
            String[] arr = args[i].split("=",2);
            // save name=value in system properties.
            System.setProperty(arr[0],arr.length > 1 ? arr[1] : "1");
        }

        int port = Integer.parseInt( 
                // connect to port passed at command line
                System.getProperty("port","12345")
                // or "12345" if there isn't one.
            );
        // create a server socket (on port, etc.)
        sock = new DatagramSocket(port);
        DatagramSocket s = sock;

        System.out.println("Started server on port "+port);

        System.out.println("Starting game thread...");
        // start game update thread.
        ShootingForDummiesServer serv = new ShootingForDummiesServer();
        Thread t = new Thread(serv);
        t.start();  // update game state (and clients).

        System.out.println("Started game thread...");

        // data buffer for our packet.
        byte[] arr = new byte[0xF000];
        for(;;){
            System.out.println("wait for packet...");

            // create a packet to write recieved data into.
            DatagramPacket p = new DatagramPacket(arr,arr.length);
            // wait (on socket) for the next packet.
            s.receive(p);

            // construct IP:port string (ie: key)
            String clientId = ""+p.getAddress()+":"+p.getPort();

            System.out.println("got packet: "+clientId);
            
            // find the client in our "database"
            Ship ci = (Ship)clients.get(clientId);
            if(ci == null){
                System.out.println("creating new ship: "+clientId);
                ci = new Ship();
                ci.id = clientId;
                ci.direction = Math.PI*2*Math.random();
                ci.x = Math.random() - 0.5;
                ci.y = Math.random() - 0.5;
                ci.ip = p.getAddress();
                ci.port = p.getPort();
                clients.put(clientId,ci);
            }else{
                System.out.println("existing ship: "+clientId);
            }
            ci.tim = System.currentTimeMillis();

            // msg from client:
            String pstr = new String(p.getData(),p.getOffset(),p.getLength()).trim();
            System.out.println("data["+clientId+":"+pstr+"]");
            String[] pstrarr = pstr.split(",");
            // ensure we have: name,color,key
            if(pstrarr.length >= 3){
                ci.name = pstrarr[0].length() > 0 ? pstrarr[0] : "Player";
                ci.color = pstrarr[1].length() > 0 ? pstrarr[1] : "0";
                // if there is a key...
                if(pstrarr[2].length() > 0){
                    int key = Integer.parseInt(pstrarr[2]);
                    if(key < 0)
                        ci.keys[-key] = false;
                    if(key > 0)
                        ci.keys[key] = true;
                }
            }
        }
    }
}

